unit hiTreeView;

interface

uses Windows,Shellapi,Kol,Share,Win;

type
  THITreeView = class(THIWin)
   private
    _Arr,ICArray,_IdxArr,CBArray:PArray;
    IList:PImageList;

    procedure _OnClick(Obj:PObj);
    function _OnBeginEdit(Sender: PControl; Item: THandle): Boolean;
    function _OnEndEdit(Sender: PControl; Item: THandle; const NewTxt: String): Boolean;
    function _OnExpanding(Sender: PControl; Item: THandle; Expand: Boolean): Boolean;
    procedure LoadFromText(Lst:PStrList);
    function IndexToHandle(Index:integer):THandle;
    function HandleToIndex(Handle:THandle):integer;

    procedure _Set(var Item:TData; var Val:TData);
    function _Get(Var Item:TData; var Val:TData):boolean;
    function _Count:integer;
    procedure _Add(var Val:TData);
    procedure _SetIcon(var Item:TData; var Val:TData);
    function _GetIcon(Var Item:TData; var Val:TData):boolean;
    procedure _AddIcon(var Val:TData);
    function  _CountIcon:integer;
    procedure SetIcons(const value:PStrListEx);
    function  ParseIconIdx(Var s:string): string;
    procedure _SetIdx(var Item: TData; var Val: TData);
    function _GetIdx(Var Item: TData; var Val: TData): boolean;
    procedure CB_Set(var Item:TData; var Val:TData);
    function CB_Get(Var Item:TData; var Val:TData):boolean;
    
   public
    _prop_Strings:string;
    _prop_Lines:boolean;
    _prop_LinesRoot:boolean;
    _prop_Tooltips:boolean;
    _prop_EditLabels:boolean;
    _prop_TrackSelect:boolean;
    _prop_CheckBoxes:boolean;
    _prop_FileName:string;
    _prop_IconByIndex:boolean;
    _prop_Delimiter:string;

    _data_FileName:THI_Event;
    _data_str:THI_Event;
    _data_Parent:THI_Event;
    _data_Child:THI_Event;
    _event_onClick:THI_Event;
    _event_onChange:THI_Event;
    _event_onFileIcon:THI_Event;
    _event_onItemParent:THI_Event;
    _event_onBeginEdit:THI_Event;
    _event_onEndEdit:THI_Event;        
    _event_onSelectChanging:THI_Event; 
    _event_onItemChild:THI_Event;
    _event_onExpand:THI_Event;
    _event_onItemPath:THI_Event;
    
    {icons}
    _data_IconID:THI_Event;
    property _prop_Icons:PStrListEx write SetIcons;

    procedure Init; override;
    destructor Destroy; override;
    procedure _work_doAdd(var _Data:TData; Index:word);
    procedure _work_doClear(var _Data:TData; Index:word);
    procedure _work_doDelete(var _Data:TData; Index:word);
    procedure _work_doLoad(var _Data:TData; Index:word);
    procedure _work_doLoadFromText(var _Data:TData; Index:word);
    procedure _work_doSave(var _Data:TData; Index:word);
    procedure _work_doSelect(var _Data:TData; Index:word);
    procedure _work_doLoadFileIcon(var _Data:TData; Index:word);
    procedure _work_doLoadPakIcons(var _Data:TData; Index:word);
    procedure _work_doItemParent(var _Data:TData; Index:word);
    procedure _work_doItemBold(var _Data:TData; Index:word);
    procedure _work_doAddChild(var _Data:TData; Index:word);    
    procedure _work_doRename(var _Data:TData; Index:word);
    procedure _work_doSort(var _Data:TData; Index:word);
    procedure _work_doItemChild(var _Data:TData; Index:word);
    procedure _work_doItemHasChild(var _Data:TData; Index:word);
    procedure _work_doItemPath(var _Data:TData; Index:word);
    procedure _var_Count(var _Data:TData; Index:word);
    procedure _var_Array(var _Data:TData; Index:word);
    procedure _var_IdxArray(var _Data:TData; Index:word);
    procedure _var_IconArray(var _Data:TData; Index:word);
    procedure _var_CheckArray(var _Data:TData; Index:word);
    procedure _var_Index(var _Data:TData; Index:word);
    procedure _var_ItemAtPos(var _Data:TData; Index:word);  
    procedure _var_ItemParent(var _Data:TData; Index:word);
    procedure _var_ItemChild(var _Data:TData; Index:word);
  end;

function MakeArrayIcon(Names:PChar; Values:array of integer):PStrListEx;

implementation

function MakeArrayIcon;begin Result := MakeArrayInteger(Names,Values);end;

function THITreeView.ParseIconIdx;
var   bmpidxerr:boolean;
      j:integer;
begin
   Result := '';
   if s = '' then Result := '-1'
   else if s[1] = '<' then begin
      delete(s,1,1);
      bmpidxerr := true;
      for j := 1 to Length(s) do begin
         if s[j] = '>' then begin
            GetTok(s,'>');
            bmpidxerr := false;
            break;
      end;
      if s[j] in ['0'..'9'] then Result := Result + s[j];
   end;
   if (bmpidxerr) or (Result = '') then Result := '-1';
   end else Result := '-1';
end;

procedure THITreeView.Init;
var Lst:PStrList;
    Fl:TTreeViewOptions;
begin

   if _prop_Lines then fl := [] else fl := [tvoNoLines];
   if not _prop_Tooltips then include(fl,tvoNoTooltips);
   if _prop_LinesRoot    then include(fl,tvoLinesRoot);
   if _prop_EditLabels   then include(fl,tvoEditLabels);
   if _prop_TrackSelect  then include(fl,tvoTrackSelect);
   if _prop_CheckBoxes   then include(fl,tvoCheckBoxes);
   if not Assigned(IList) then begin  
      IList:= NewImageList(nil);    
      IList.BkColor := clWindow;
      IList.ImgWidth := 16;
      IList.ImgHeight := 16;
   end;
   Control := NewTreeView(FParent,fl,IList,nil);
   Control.OnTVExpanding := _OnExpanding;
   Control.OnSelChange := _OnClick;
   Control.OnTVBeginEdit := _OnBeginEdit;
   Control.OnTVEndEdit := _OnEndEdit;
   Lst := NewStrList;
   Lst.Text := _prop_Strings;
   LoadFromText(Lst);
   inherited;
end;

destructor THITreeView.Destroy;
begin
   if Assigned(IList) then IList.free;
   if _Arr <> nil then dispose(_Arr);
   if ICArray <> nil then dispose(ICArray);
   if _IdxArr <> nil then dispose(_IdxArr);
   if CBArray <> nil then Dispose(CBArray);
   inherited;
end;

procedure THITreeView.LoadFromText;
var   i:smallint;
      ind:smallint;
      Last,Prn:cardinal;
      s:string;
begin
   Control.Clear;
   Last := 0;
   Prn := 0;
   for i := 0 to Lst.Count-1 do begin
      s := Lst.Items[i];
      if s = '(' then
         Prn := Last
      else if s = ')' then begin
         last := Prn;
         Prn := Control.TVItemParent[Prn];
      end else begin
         if not _prop_IconByIndex then begin
            Last := Control.TVInsert(Prn,Control.Count-1,s);
            ind := Control.Count-1; 
         end else begin
            ind := str2int(ParseIconIdx(s));
            Last := Control.TVInsert(Prn,Control.Count-1,s);
         end;
         Control.TVItemImage[Last] := ind;
         Control.TVItemSelImg[Last] := Control.TVItemImage[Last];
      end;
   end;
   Lst.Free;
   _hi_OnEvent(_event_onChange);
end;

procedure THITreeView._OnClick;
begin
   _hi_OnEvent(_event_onClick,Control.TVItemText[Control.TVSelected])
end;

function THITreeView._OnBeginEdit;
begin
   Result := true;
   _hi_OnEvent(_event_onBeginEdit, Control.TVItemText[Item]);
end;

function THITreeView._OnEndEdit;
begin
   Result := true;
   _hi_OnEvent(_event_onEndEdit, NewTxt)
end;

function THITreeView._OnExpanding;
begin
   Result := true;
   if Expand then _hi_OnEvent(_event_onExpand, HandleToIndex(Item));
end;

function THITreeView.IndexToHandle;
var   i:smallint;
      p:THandle;
begin
   p := 0;
   i := 0;
   while i <= Index do begin
      if Control.TVItemChild[p] > 0 then
         p := Control.TVItemChild[p]
      else if Control.TVItemNext[p] > 0 then
         p := Control.TVItemNext[p]
      else begin
         repeat
            p := Control.TVItemParent[p];
         until (p = 0)or(Control.TVItemNext[p] > 0);
         if p > 0 then
            p := Control.TVItemNext[p];
      end;
      inc(i);
   end;
   Result := p;
end;

function THITreeView.HandleToIndex;
var   ind:integer;

      function HTI(prn:THandle):boolean;
      begin
         repeat
            if prn = Handle then
               Result := true
            else if prn = 0 then
               Result := false
            else if Control.TVItemChild[prn] > 0 then
               Result := HTI( Control.TVItemChild[prn] )
            else
               Result := false;
            inc(ind);
            prn := Control.TVItemNext[prn];
         until Result or( prn = 0 );
      end;

begin
   ind := 0;
   if HTI(Control.TVRoot) then
      Result := Ind-1
   else
      Result := -1;
end;

procedure THITreeView._work_doAdd;
var   item:THandle;
      s:string;
      n:smallint;
      dt:TData;
begin
   s := ReadString(_Data,_data_str,'');
   dt := ReadData(_Data,_data_IconID);
   if _prop_IconByIndex then begin
      if _IsNull(dt) then
         n := str2int(ParseIconIdx(s))
      else
         n := ToInteger(dt);
   end else
      n := ToInteger(dt);
   item := Control.TVInsert(Control.TVSelected, 0, s);
   Control.TVItemImage[item] := n;
   Control.TVItemSelImg[item] := n;
   _hi_OnEvent(_event_onChange);
end;

procedure THITreeView._work_doClear;
begin
   Control.Clear;
   _hi_OnEvent(_event_onChange);
end;

procedure THITreeView._work_doDelete;
begin
   Control.TVDelete(IndexToHandle(ToInteger(_Data)));
   _hi_OnEvent(_event_onChange);
end;

procedure THITreeView._work_doLoad;
var   Lst:PStrList;
begin
   Lst := NewStrList;
   Lst.LoadFromFile(ReadFileName(ReadString(_Data,_data_FileName,_prop_FileName)));
   LoadFromText(Lst);
end;

procedure THITreeView._work_doLoadFromText;
var   Lst:PStrList;
begin
   Lst := NewStrList;
   Lst.Text := ReadString(_Data,_data_FileName,_prop_FileName);
   LoadFromText(Lst);
end;

procedure THITreeView._work_doSave;
var   Lst:PStrList;

      procedure Save(prn:THandle);
      var   sidx:string;
            idx:smallint;
      begin
         if prn > 0 then begin
            sidx := '';
            if _prop_IconByIndex then begin
               idx := Control.TVItemImage[prn];
               if idx >= 0 then sidx := '<' + int2str(idx) + '>'
            end;   
            Lst.Add( sidx + Control.TVItemText[prn] );
            if Control.TVItemChild[prn] > 0 then begin
               Lst.Add('(');
               Save(Control.TVItemChild[prn]);
               Lst.Add(')');
            end;
            Save( Control.TVItemNext[prn] );

         end;
      end;

begin
   Lst := NewStrList;
   Save(Control.TVRoot);
   Lst.SaveToFile(ReadFileName(ReadString(_Data,_data_FileName,_prop_FileName)));
   Lst.Free;
end;

procedure THITreeView._work_doSelect;
begin
   Control.TVSelected := IndexToHandle( ToInteger(_Data) );
end;

procedure THITreeView._work_doRename;
begin
   Control.TVItemText[ Control.TVSelected ] := ToString(_Data);
end;

procedure THITreeView._var_Count;
begin
   dtInteger(_Data,Control.Count);
end;

procedure THITreeView._Set(var Item:TData; var Val:TData);
var   ind:integer;
      s:string;
      n:integer;
begin
   ind := ToInteger(Item);
   if (ind >= 0 ) and (ind < Control.Count ) then begin
      s := ToString(Val);
      if _prop_IconByIndex then begin
         n := str2int(ParseIconIdx(s));
         Control.TVItemImage[IndexToHandle(ind)] := n;
         Control.TVItemSelImg[IndexToHandle(ind)] := n;         
      end;
   end;
   Control.TVItemText[ IndexToHandle(ind) ] := s;
   Control.Invalidate;
end;

function THITreeView._Get(Var Item:TData; var Val:TData):boolean;
var   ind:integer;
begin
   ind := ToInteger(Item);
   if(ind >= 0 )and(ind < Control.Count )then begin
      dtString(Val,Control.TVItemText[ IndexToHandle(ind) ]);
      Result := true;
   end else
      Result := false;
end;

function THITreeView._Count:integer;
begin
   Result := Control.Count;
end;

procedure THITreeView._Add(var Val:TData);
begin
   _work_doAdd(Val,0);
end;

procedure THITreeView._var_Array;
begin
   if _Arr = nil then
      _Arr := CreateArray(_Set,_Get,_Count,_Add);
   dtArray(_Data,_Arr);
end;

procedure THITreeView._var_Index;
begin
   dtInteger(_Data,HandleToIndex(Control.TVSelected));
end;

procedure THITreeView.SetIcons;
var   i:integer;
begin
   IList := NewImageList(Control);
   IList.BkColor := clWindow;
   IList.ImgWidth := 16;
   IList.ImgHeight := 16;
   for i := 0 to value.Count-1 do
      IList.AddIcon(Value.Objects[i]);
end;

//-------------------         ---------------------

//IconArray -  
//
procedure THITreeView._var_IconArray;
begin
   if not Assigned(ICArray) then ICArray := CreateArray(_SetIcon,_GetIcon,_CountIcon,_AddIcon);
   dtArray(_Data,ICArray);
end;

procedure THITreeView._SetIcon(var Item:TData; var Val:TData);
var   ind:integer;
begin
   IList.BkColor := clWindow;
   IList.ImgWidth := 16;
   IList.ImgHeight := 16;
   ind:= ToIntIndex(Item);
   if (ind >= 0) and (ind < IList.Count) and _IsIcon(Val) then
      IList.ReplaceIcon(ind, ToIcon(val).handle);
   Control.Invalidate;
end;

function THITreeView._GetIcon(Var Item:TData; var Val:TData):boolean;
begin
   Result:= False;
end;

procedure THITreeView._AddIcon(var Val:TData);
begin
   IList.BkColor := clWindow;
   IList.ImgWidth := 16;
   IList.ImgHeight := 16;
   if _IsIcon(Val) then IList.AddIcon(ToIcon(Val).Handle);
   Control.Invalidate;
end;

function THITreeView._CountIcon:integer;
begin
   Result:= IList.Count;
end;

//ItemAtPos -     ,    MouseX, MouseY
//
procedure THITreeView._var_ItemAtPos;
var   where:dword;
begin
   dtInteger(_Data, HandleToIndex(Control.TVItemAtPos(Ms.X, Ms.Y, where)));
end;

//doLoadFileIcon -  ,   .     
//
procedure THITreeView._work_doLoadFileIcon;
const Flags = SHGFI_ICON or SHGFI_ICONLOCATION or SHGFI_SMALLICON or SHGFI_TYPENAME or SHGFI_SYSICONINDEX;
var   ico: PIcon;
      fn:string;
      dt:TData;
      SFI: TShFileInfo;      
begin
   fn := ToString(_Data);
   ico:= NewIcon;
   ShGetFileInfo(PChar(fn), 0, SFI, SizeOf(SFI), Flags);
   ico.handle:= SFI.hIcon; 
   if ico.Handle <> 0 then dtIcon(dt,ico) else dtNull(dt); 
   _hi_onEvent(_event_onFileIcon, dt);
   ico.free;
end;

//doLoadPakIcons -      (*.exe,*.dll,*.ocx,*.icl)   
//
procedure THITreeView._work_doLoadPakIcons;
var   i, IconCount:integer;
      fn:string;
begin
   fn:= ToString(_Data);
   IconCount:= GetFileIconCount(fn);
   if IconCount <> 0 then begin
      i:= 0;
      repeat
         IList.BkColor := clWindow;
         IList.ImgWidth := 16;
         IList.ImgHeight := 16;
         IList.AddIcon(ExtractIcon(hInstance, PChar(fn), i));
         inc(i);
      until i = IconCount;
   end;
   Control.Invalidate;
end;

//doItemParent -           
//
procedure THITreeView._work_doItemParent;
var   idx:integer;
begin
   idx := HandleToIndex(Control.TVItemParent[IndexToHandle(ToInteger(_Data))]);
   _hi_OnEvent(_event_onItemParent, idx);
end;

//ItemParent -            Child
//
procedure THITreeView._var_ItemParent;
var   idx:integer;
begin
   idx := HandleToIndex(Control.TVItemParent[IndexToHandle(ToIntegerEvent(_data_Child))]);
   dtInteger(_Data, idx)
end;

//AddChild -  ,      Parent
//
procedure THITreeView._work_doAddChild;
var   item:THandle;
      s:string;
      n:smallint;
      dt:TData;
begin
   s := ReadString(_Data,_data_str,'');
   dt := ReadData(_Data,_data_IconID);
   if _prop_IconByIndex then begin
      if _IsNull(dt) then
         n := str2int(ParseIconIdx(s))
      else
         n := ToInteger(dt);
   end else
      n := ToInteger(dt);
   item := Control.TVInsert(IndexToHandle(ReadInteger(_Data,_data_Parent)), 0, s);
   Control.TVItemImage[item] := n;
   Control.TVItemSelImg[item] := n;
   _hi_OnEvent(_event_onChange);
end;

//doItemBold - /        
//
procedure THITreeView._work_doItemBold;
var   item:THandle;
begin
   item := IndexToHandle(ToInteger(_Data)); 
   Control.TVItemBold[item] := not Control.TVItemBold[item];
end;

//IdxArray -    
//
procedure THITreeView._SetIdx(var Item:TData; var Val:TData);
var   ind:integer;
      idx:smallint;
begin
   ind := ToIntIndex(Item);
   idx := ToInteger(Val);
   if idx < 0 then idx := -1; 
   if (ind >= 0) and (ind < Control.Count) then
      Control.TVItemImage[IndexToHandle(ind)]:= idx;
   Control.Invalidate;
end;

function THITreeView._GetIdx(Var Item:TData; var Val:TData):boolean;
var   ind:integer;
      idx:smallint;
begin
   ind := ToIntIndex(Item);
   idx := Control.TVItemImage[IndexToHandle(ind)]; 
   if (ind >= 0 ) and (ind < Control.Count) then begin
      dtInteger(Val, idx);
      Result := true;
   end else
      Result := false;
end;

procedure THITreeView._var_IdxArray;
begin
   if _IdxArr = nil then
      _IdxArr := CreateArray(_SetIdx,_GetIdx,_Count,nil);
   dtArray(_Data, _IdxArr);
end;

//CheckArray -    (0 -  , 1 - )
//
procedure THITreeView._var_CheckArray;
begin
   if not Assigned(CBArray) then
      CBArray := CreateArray(CB_Set,CB_Get,_Count,nil);
   dtArray(_Data,CBArray);
end;

procedure THITreeView.CB_Set(var Item:TData; var Val:TData);
var   ind:integer;
begin
   ind:= ToIntIndex(Item);
   if not _prop_CheckBoxes then exit;
   if (ind >= 0) and (ind < Control.Count) then
      Control.TVItemStateImg[IndexToHandle(ind)] := ToInteger(Val) + 1;
end;

function THITreeView.CB_Get(Var Item:TData; var Val:TData):boolean;
var   ind:integer;
begin
   Result:= False;
   if not _prop_CheckBoxes then exit;
   ind:= ToIntIndex(Item);
   if (ind >= 0) and (ind < Control.Count) then begin
      dtInteger(Val, Control.TVItemStateImg[IndexToHandle(ind)] - 1);
      Result:= True;
   end;
end;

//doItemHasChild -       
//
procedure THITreeView._work_doItemHasChild;
begin
   Control.TVItemHasChildren[IndexToHandle(ToInteger(_Data))] := true;
end;

//doSort -   
//
procedure THITreeView._work_doSort;
begin
   Control.TVSort(0);
end;

//doItemChild -            
//
procedure THITreeView._work_doItemChild;
var   idx:integer;
begin
   idx := HandleToIndex(Control.TVItemChild[IndexToHandle(ToInteger(_Data))]);
   _hi_OnEvent(_event_onItemChild, idx);
end;

//ItemChild -            Parent
//
procedure THITreeView._var_ItemChild;
var   idx:integer;
begin
   idx := HandleToIndex(Control.TVItemChild[IndexToHandle(ToIntegerEvent(_data_Parent))]);
   dtInteger(_Data, idx)
end;

//ItemPath -  ""   Delimiter
//                    
//
procedure THITreeView._work_doItemPath;
var   idx:integer;
      s:string;
begin
   idx := ToInteger(_Data);
   if _prop_Delimiter[1] = '\' then s := '\' else s := '';  
   _hi_OnEvent(_event_onItemPath, Control.TVItemPath(IndexToHandle(idx), _prop_Delimiter[1]) + s); 
end;

end.
